# Copyright (c) 2020 Seven Bridges
# Given a cwl dict, if it is a workflow, split out any inlined steps into their own files

from typing import List
import pathlib
import sys

import ruamel.yaml

from .version import __version__
from .formatter import stringify_dict, leading_comment_lines


yaml = ruamel.yaml.YAML()


class CWLProcess:
    def __init__(self, cwl: dict, file_path: pathlib.Path):
        self.cwl = cwl
        self.file_path = file_path

    def __str__(self):
        return stringify_dict(self.cwl)

    def save(self):
        self.file_path.parent.mkdir(parents=True, exist_ok=True)
        self.file_path.write_text(leading_comment_lines("") + stringify_dict(self.cwl))


def explode(cwl: CWLProcess) -> List[CWLProcess]:
    _processes = [cwl]
    _cwl = cwl.cwl
    if _cwl.get("class") == "Workflow":
        _cwl_steps = _cwl.get("steps", {})
        _is_dict = isinstance(_cwl_steps, dict)
        for _k, _step in (_cwl_steps.items() if _is_dict else enumerate(_cwl_steps)):
            _step_id = _k if _is_dict else _step.get("id")
            if _step_id is not None:
                _run = _step.get("run")
                if isinstance(_run, dict):
                    step_path = \
                        cwl.file_path.parent / \
                        (cwl.file_path.name + ".steps") / \
                        (_step_id + ".cwl")
                    _step["run"] = str(step_path.relative_to(cwl.file_path.parent))
                    _processes += explode(CWLProcess(_run, step_path))

    return _processes


def main():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=f"Rabix/cwl-explode v{__version__}\n"
                    "Explodes CWL workflow with inlined steps")
    parser.add_argument("cwlfile")
    parser.add_argument("outname")
    args = parser.parse_args()

    fp = pathlib.Path(args.cwlfile).absolute()
    as_dict = yaml.load(fp.read_text())
    fp_out = pathlib.Path(args.outname).absolute()
    for n, exploded in enumerate(explode(CWLProcess(as_dict, fp_out))):
        sys.stderr.write(f"{n + 1}: {exploded.file_path.relative_to(fp.parent)}\n")
        exploded.save()


if __name__ == "__main__":
    main()



# Copyright (c) 2020 Seven Bridges

import pathlib

from cwlformat.formatter import cwl_format, yaml, format_node


current_path = pathlib.Path(__file__).parent


def test_formatting_battery():

    path = current_path / "cwl"
    for raw_name in path.glob("original-*"):
        expected_name = raw_name.parent / pathlib.Path("formatted-" + "-".join(raw_name.stem.split("-")[1:]) + ".cwl")
        formatted_cwl = cwl_format(raw_name.open("r").read())
        expected_raw_cwl = expected_name.open("r").read()

        assert formatted_cwl == expected_raw_cwl


def test_node_conservation():
    path = current_path / "cwl"
    for raw_name in path.glob("original-*"):
        original_cwl = yaml.load(raw_name.open("r").read())
        formatted_cwl = format_node(original_cwl, node_path=[])

        assert formatted_cwl == original_cwl



# Copyright (c) 2020 Seven Bridges

from typing import Union
import sys
import pathlib
try:
    from importlib.resources import read_text
except ImportError:
    # Python 3.6 fallback: https://importlib-resources.readthedocs.io/en/latest/
    from importlib_resources import read_text

import ruamel.yaml
from ruamel.yaml import scalarstring
from ruamel.yaml.compat import StringIO
from ruamel.yaml.comments import CommentedMap

from cwlformat.version import __version__

yaml = ruamel.yaml.YAML()
yaml.indent(mapping=2, sequence=2, offset=0)
Literal = ruamel.yaml.scalarstring.LiteralScalarString

key_order_dict = yaml.load(read_text("cwlformat", "keyorder.yml"))

hash_bang = "#!/usr/bin/env cwl-runner\n\n"
hash_bang_pre = "#!/usr/bin/env "


def leading_comment_lines(raw_cwl: str):

    top_comment = []

    if len(raw_cwl) > 0 and raw_cwl.lstrip()[0] != "{":
        for _line in raw_cwl.splitlines(keepends=True):
            line = _line.strip()
            if line == "" or line[0] == "#":
                top_comment += [_line]
            else:
                break

    if len(top_comment) == 0 or not top_comment[0].startswith(hash_bang_pre):
        top_comment = [hash_bang] + top_comment

    return "".join(top_comment)


def format_node(cwl: Union[dict, list, str], node_path=None):
    if isinstance(cwl, str):
        if len(cwl) > 80:
            return Literal(cwl)
        else:
            return cwl

    elif isinstance(cwl, dict):
        _fmt_cwl = CommentedMap([
            (k, format_node(v, node_path + [k])) for k, v in reorder_node(cwl, node_path)])
        if _fmt_cwl.get("class") in ["CommandLineTool", "ExpressionTool", "Workflow"]:
            add_space_between_main_sections(_fmt_cwl)
        return _fmt_cwl

    elif isinstance(cwl, list):
        return [format_node(v, node_path) for v in cwl]

    else:
        return cwl


def add_space_between_main_sections(cwl: CommentedMap):
    for k in cwl.keys():
        if k in ["inputs", "outputs", "steps", "requirements", "hints", "baseCommand"]:
            cwl.yaml_set_comment_before_after_key(key=k, before="\n")


def reorder_node(cwl: dict, node_path: list) -> dict:
    known_key_order = key_order_dict.get(
        infer_type(cwl, node_path), key_order_dict["generic-ordering"])
    extra_keys = sorted(set(cwl.keys()) - set(known_key_order))

    for k in known_key_order + extra_keys:
        if k in cwl:
            yield k, cwl[k]


def infer_type(cwl: dict, node_path: list):
    if "class" in cwl:
        return cwl["class"]
    else:
        return "generic-ordering"


def cwl_format(raw_cwl: str) -> str:
    as_dict = yaml.load(raw_cwl)
    return leading_comment_lines(raw_cwl) + stringify_dict(as_dict)


def stringify_dict(as_dict: dict) -> str:
    as_dict = format_node(as_dict, node_path=[])
    stream = StringIO()
    yaml.dump(as_dict, stream)
    return stream.getvalue()


def main():
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=f"Rabix/cwl-format v{__version__}\n"
                    "A very opinionated code formatter for CWL")
    parser.add_argument("cwlfile")
    parser.add_argument("--inplace", action="store_true",
                        help="Instead of writing formatted code to stdout, overwrite original file")
    args = parser.parse_args()
    fp = pathlib.Path(args.cwlfile)
    formatted = cwl_format(fp.read_text())
    if args.inplace:
        fp.write_text(formatted)
    else:
        sys.stdout.write(formatted)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import pathlib
from setuptools import setup, find_packages

current_path = pathlib.Path(__file__).parent
ver_path = pathlib.Path(current_path, "cwlformat", "version.py")
_ver = {}
exec(ver_path.open("r").read(), _ver)
version = _ver["__version__"]

readme = pathlib.Path(current_path, "Readme.md").read_text()

setup(
    name='cwlformat',
    python_requires='>=3.6.0',
    version=version,
    description='A prettifier for CWL code',
    long_description=readme,
    long_description_content_type="text/markdown",
    author='Kaushik Ghose',
    author_email='kaushik.ghose@sbgenomics.com',
    url='https://github.com/rabix/cwl-format',
    packages=find_packages(exclude=('tests', 'docs')),
    entry_points={
        'console_scripts': [
            'cwl-format=cwlformat.formatter:main',
            'cwl-explode=cwlformat.explode:main'
        ],
    },
    include_package_data=True,
    install_requires=[
        "ruamel.yaml >= 0.16.12",
    ],
    extras_require={':python_version<"3.7"': ['importlib-resources']},
)



# Copyright (c) 2020 Seven Bridges

import pathlib

import ruamel.yaml

from cwlformat.explode import explode, CWLProcess

yaml = ruamel.yaml.YAML()

current_path = pathlib.Path(__file__).parent


def test_explode():
    path = current_path / "cwl"
    src_fp = path / "formatted-atac-seq-pipeline.cwl"
    fp_out = src_fp.parent / "expected-exploded-atac-seq.cwl"

    as_dict = yaml.load(src_fp.read_text())
    for exploded in explode(CWLProcess(as_dict, fp_out)):
        assert exploded.file_path.exists()
